//
//	Adler32
//

#ifndef		_ADLER32_INC_
#define		_ADLER32_INC_

class Adler32
{
public:
	Adler32()
	{
		init();
	}
	unsigned long result() const
	{
		return (unsigned long)s2_ << 16 | (unsigned long)s1_;
	}
	Adler32 & init()
	{
		s1_ = 1;
		s2_ = 0;
		return *this;
	}
	Adler32 & put(unsigned char x)
	{
		s1_ = (s1_ +  x ) % 65521;
		s2_ = (s2_ + s1_) % 65521;
		return *this;
	}
	Adler32 & puts(const void *p, unsigned long len)
	{
		for (unsigned long i = 0; i < len; i++)
			put(((const unsigned char *)p)[i]);
		return *this;
	}
	Adler32 & puts(const char *s)
	{
		for ( ; *s; s++)
			put(*s);
		return *this;
	}
	Adler32 & operator () ()
	{
		return init();
	}
	Adler32 & operator << (unsigned char x)
	{
		return put(x);
	}
	Adler32 & operator << (const char *s)
	{
		return puts(s);
	}
	operator unsigned long () const
	{
		return result();
	}
protected:
	unsigned short s1_, s2_;
};

#endif	//	_ADLER32_INC_
