//
//	CRC32
//

#ifndef		_CRC32_INC_
#define		_CRC32_INC_

class CRC32
{
public:
	CRC32(unsigned long i = 0)
	{
		s_table_init();
		init(i);
	}
	unsigned long result() const
	{
		return ~crc_;
	}
	CRC32 & init(unsigned long i = 0)
	{
		crc_ = ~i;
		return *this;
	}
	CRC32 & put(unsigned char x)
	{
		crc_ = s_table_[(crc_ ^ x) & 0xFF] ^ (crc_ >> 8);
		return *this;
	}
	CRC32 & puts(const void *p, unsigned long n)
	{
		for (const unsigned char *ptr = (const unsigned char *)p,
		     *end = ptr + n; ptr < end; ptr++) {
			put(*ptr);
		}
		return *this;
	}
	CRC32 & puts(const char *s)
	{
		for ( ; *s; s++)
			put(*s);
		return *this;
	}
	CRC32 & operator = (unsigned long i)
	{
		return init(i);
	}
	CRC32 & operator << (unsigned char x)
	{
		return put(x);
	}
	CRC32 & operator << (const char *s)
	{
		return puts(s);
	}
	operator unsigned long () const
	{
		return result();
	}
	virtual ~CRC32()
	{
	}
protected:
	unsigned long crc_;
private:
	static void s_table_init()
	{
		if (s_table_[1] != 0)
			return;
		const unsigned long poly = 0xEDB88320;
        for (unsigned long i = 0; i < 256; i++) {
			unsigned long c = i;
			for (unsigned long k = 0; k < 8; k++)
				c = c & 1 ? poly ^ (c >> 1) : c >> 1;
			s_table_[i] = c;
        }
	}
	static unsigned long s_table_[256];
};
unsigned long CRC32::s_table_[256] = {0};

#endif	//	_CRC32_INC_
