//
//	Property
//

#ifndef		_PROPERTY_INC_
#define		_PROPERTY_INC_

#include <assert.h>

template <class T, class C>
class property
{
public:
	operator T () const
	{
		#ifndef NDEBUG
			assert(access_ & GET);
		#endif
		return (obj_->*get_)();
	}
	T operator = (T const &x)
	{
		#ifndef NDEBUG
			assert(access_ & SET);
		#endif
		return (obj_->*set_)(x);
	}
	T operator = (property<T, C> const &x)
	{
		#ifndef NDEBUG
			assert(access_ & SET);
		#endif
		return (obj_->*set_)((T)x);
	}
	bool operator == (property<T, C> const &x)
	{
		return (T)*this == (T)x;
	}
	bool operator != (property<T, C> const &x)
	{
		return (T)*this != (T)x;
	}
public:
	void operator () (C *obj, T (C::*get)() const, T (C::*set)(T const &))
	{
		assert(!obj_);
		obj_ = obj;
		get_ = get;
		set_ = set;
		#ifndef NDEBUG
			access_ = Access(GET | SET);
		#endif
	}
	void operator () (C *obj, T (C::*get)() const)
	{
		assert(!obj_);
		obj_ = obj;
		get_ = get;
		#ifndef NDEBUG
			access_ = GET;
		#endif
	}
	void operator () (C *obj, T (C::*set)(T const &))
	{
		assert(!obj_);
		obj_ = obj;
		set_ = set;
		#ifndef NDEBUG
			access_ = SET;
		#endif
	}
public:
	property()
		: obj_(0)
	{
		#ifndef NDEBUG
			access_ = NONE;
		#endif
	}
	property(C *obj, T (C::*get)() const, T (C::*set)(T const &))
		: obj_(0)
	{
		(*this)(obj, get, set);
	}
	property(C *obj, T (C::*get)() const)
		: obj_(0)
	{
		(*this)(obj, get);
	}
	property(C *obj, T (C::*set)(T const &))
		: obj_(0)
	{
		(*this)(obj, set);
	}
protected:
	C *obj_;
	T (C::*get_)() const;
	T (C::*set_)(T const &);
	#ifndef NDEBUG
		enum Access { NONE = 0, GET = 1, SET = 2 } access_;
	#endif
};

template <class T, class C, class P>
class property_param
{
public:
	operator T () const
	{
		#ifndef NDEBUG
			assert(access_ & GET);
		#endif
		return (obj_->*get_)(par_);
	}
	T operator = (T const &x)
	{
		#ifndef NDEBUG
			assert(access_ & SET);
		#endif
		return (obj_->*set_)(x, par_);
	}
	T operator = (property_param<T, C, P> const &x)
	{
		#ifndef NDEBUG
			assert(access_ & SET);
		#endif
		return (obj_->*set_)((T)x, par_);
	}
	bool operator == (property_param<T, C, P> const &x)
	{
		return (T)*this == (T)x;
	}
	bool operator != (property_param<T, C, P> const &x)
	{
		return (T)*this != (T)x;
	}
public:
	void operator () (C *obj, T (C::*get)(P const &) const, T (C::*set)(T const &, P const &), P const &par)
	{
		assert(!obj_);
		obj_ = obj;
		get_ = get;
		set_ = set;
		par_ = par;
		#ifndef NDEBUG
			access_ = Access(GET | SET);
		#endif
	}
	void operator () (C *obj, T (C::*get)(P const &) const, P const &par)
	{
		assert(!obj_);
		obj_ = obj;
		get_ = get;
		par_ = par;
		#ifndef NDEBUG
			access_ = GET;
		#endif
	}
	void operator () (C *obj, T (C::*set)(T const &, P const &), P const &par)
	{
		assert(!obj_);
		obj_ = obj;
		set_ = set;
		par_ = par;
		#ifndef NDEBUG
			access_ = SET;
		#endif
	}
public:
	property_param()
		: obj_(0)
	{
		#ifndef NDEBUG
			access_ = NONE;
		#endif
	}
	property_param(C *obj, T (C::*get)(P const &) const, T (C::*set)(T const &, P const &), P const &par)
		: obj_(0)
	{
		(*this)(obj, get, set, par);
	}
	property_param(C *obj, T (C::*get)(P const &) const, P const &par)
		: obj_(0)
	{
		(*this)(obj, get, par);
	}
	property_param(C *obj, T (C::*set)(T const &, P const &), P const &par)
		: obj_(0)
	{
		(*this)(obj, set, par);
	}
protected:
	C *obj_;
	T (C::*get_)(P const &) const;
	T (C::*set_)(T const &, P const &);
	P par_;
	#ifndef NDEBUG
		enum Access { NONE = 0, GET = 1, SET = 2 } access_;
	#endif
};

template <class T, class C, class I>
class property_array
{
public:
	T & operator [] (I const &i)
	{
		#ifndef NDEBUG
			assert(access_ & ARR);
		#endif
		return (obj_->*arr_)(i);
	}
	operator T * () const
	{
		#ifndef NDEBUG
			assert(access_ & GET);
		#endif
		return (obj_->*get_)();
	}
	T * operator = (T *x)
	{
		#ifndef NDEBUG
			assert(access_ & SET);
		#endif
		return (obj_->*set_)(x);
	}
	T * operator = (property_param<T, C, I> const &x)
	{
		#ifndef NDEBUG
			assert(access_ & SET);
		#endif
		return (obj_->*set_)((T *)x);
	}
public:
	void operator () (C *obj, T & (C::*arr)(I const &), T * (C::*get)() const, T * (C::*set)(T *))
	{
		assert(!obj_);
		obj_ = obj;
		arr_ = arr;
		get_ = get;
		set_ = set;
		#ifndef NDEBUG
			access_ = Access(ARR | GET | SET);
		#endif
	}
	void operator () (C *obj, T & (C::*arr)(I const &), T * (C::*get)() const)
	{
		assert(!obj_);
		obj_ = obj;
		arr_ = arr;
		get_ = get;
		#ifndef NDEBUG
			access_ = Access(ARR | GET);
		#endif
	}
	void operator () (C *obj, T & (C::*arr)(I const &), T * (C::*set)(T *))
	{
		assert(!obj_);
		obj_ = obj;
		arr_ = arr;
		set_ = set;
		#ifndef NDEBUG
			access_ = Access(ARR | SET);
		#endif
	}
	void operator () (C *obj, T & (C::*arr)(I const &))
	{
		assert(!obj_);
		obj_ = obj;
		arr_ = arr;
		#ifndef NDEBUG
			access_ = ARR;
		#endif
	}
public:
	property_array()
		: obj_(0)
	{
		#ifndef NDEBUG
			access_ = NONE;
		#endif
	}
	property_array(C *obj, T & (C::*arr)(I const &), T * (C::*get)() const, T * (C::*set)(T *))
		: obj_(0)
	{
		(*this)(obj, arr, get, set);
	}
	property_array(C *obj, T & (C::*arr)(I const &), T * (C::*get)() const)
		: obj_(0)
	{
		(*this)(obj, arr, get);
	}
	property_array(C *obj, T & (C::*arr)(I const &), T * (C::*set)(T *))
		: obj_(0)
	{
		(*this)(obj, arr, set);
	}
	property_array(C *obj, T & (C::*arr)(I const &))
		: obj_(0)
	{
		(*this)(obj, arr);
	}
protected:
	C *obj_;
	T & (C::*arr_)(I const &);
	T * (C::*get_)() const;
	T * (C::*set_)(T *);
	#ifndef NDEBUG
		enum Access { NONE = 0, GET = 1, SET = 2, ARR = 4 } access_;
	#endif
};

#endif	//	_PROPERTY_INC_
